@@ -1,12 +1,13 @@
-/* @(#)acl_unix.c	1.12 03/01/21 Copyright 2001 J. Schilling */
+/* @(#)acl_unix.c	1.40 09/07/11 Copyright 2001-2009 J. Schilling */
+#include <schily/mconfig.h>
 #ifndef lint
-static	char sccsid[] =
-	"@(#)acl_unix.c	1.12 03/01/21 Copyright 2001 J. Schilling";
+static	UConst char sccsid[] =
+	"@(#)acl_unix.c	1.40 09/07/11 Copyright 2001-2009 J. Schilling";
 #endif
 /*
  *	ACL get and set routines for unix like operating systems.
  *
- *	Copyright (c) 2001 J. Schilling
+ *	Copyright (c) 2001-2009 J. Schilling
  *
  *	This implementation currently supports POSIX.1e and Solaris ACLs.
  *	Thanks to Andreas Gruenbacher <ag@bestbits.at> for the first POSIX ACL
@@ -18,23 +19,27 @@ static	char sccsid[] =
  *	on true64.
  */
 /*
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
- * any later version.
+ * The contents of this file are subject to the terms of the
+ * Common Development and Distribution License, Version 1.0 only
+ * (the "License").  You may not use this file except in compliance
+ * with the License.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * See the file CDDL.Schily.txt in this distribution for details.
  *
- * You should have received a copy of the GNU General Public License along with
- * this program; see the file COPYING.  If not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * When distributing Covered Code, include this CDDL HEADER in each
+ * file and include the License file CDDL.Schily.txt from this distribution.
  */
 
-#include <mconfig.h>
+
+#include <schily/mconfig.h>
 #ifdef	USE_ACL
+
+#ifdef	OWN_ACLTEXT
+#if	defined(UNIXWARE) && defined(HAVE_ACL)
+#	define	HAVE_SUN_ACL
+#	define	HAVE_ANY_ACL
+#endif
+#endif
 /*
  * HAVE_ANY_ACL currently includes HAVE_POSIX_ACL and HAVE_SUN_ACL.
  * This definition must be in sync with the definition in star_unix.c
@@ -48,28 +53,50 @@ static	char sccsid[] =
 #	endif
 #endif
 
-#ifdef	USE_ACL
-#include <stdio.h>
-#include <errno.h>
+#include <schily/stdio.h>
+#include <schily/errno.h>
 #include "star.h"
 #include "props.h"
 #include "table.h"
-#include <standard.h>
-#include <stdxlib.h>	/* Needed for Solaris ACL code (malloc/free) */
-#include <unixstd.h>
-#include <dirdefs.h>
-#include <strdefs.h>
-#include <statdefs.h>
-#include <schily.h>
+#include <schily/standard.h>
+#include <schily/stdlib.h>	/* Needed for Solaris ACL code (malloc/free) */
+#include <schily/unistd.h>
+#include <schily/dirent.h>
+#include <schily/string.h>
+#include <schily/stat.h>
+#include <schily/schily.h>
+#include <schily/idcache.h>
 #include "starsubs.h"
+#include "checkerr.h"
+
+#ifdef	USE_ACL
 
 #ifdef	HAVE_SYS_ACL_H
 #	include <sys/acl.h>
 #endif
 
+/*
+ * Define some things that either are missing or defined in a different way
+ * on SCO UnixWare
+ */
+#if	defined(UNIXWARE)
+typedef struct acl	aclent_t;
+#endif
+#ifndef	GETACL
+#define	GETACL	ACL_GET
+#endif
+#ifndef	SETACL
+#define	SETACL	ACL_SET
+#endif
+#ifndef	GETACLCNT
+#define	GETACLCNT	ACL_CNT
+#endif
+#ifndef	MIN_ACL_ENTRIES
+#define	MIN_ACL_ENTRIES	NACLBASE
+#endif
+
 #define	ROOT_UID	0
 
-extern	int	uid;
 extern	BOOL	nochown;
 extern	BOOL	numeric;
 
@@ -82,26 +109,35 @@ extern	BOOL	numeric;
 LOCAL char acl_access_text[PATH_MAX+1];
 LOCAL char acl_default_text[PATH_MAX+1];
 
+EXPORT	void	opt_acl		__PR((void));
 EXPORT	BOOL	get_acls	__PR((FINFO *info));
 EXPORT	void	set_acls	__PR((FINFO *info));
 
 #ifdef	HAVE_POSIX_ACL
 LOCAL	BOOL	acl_to_info	__PR((char *name, int type, char *acltext));
-LOCAL	BOOL	acl_add_ids	__PR((char *infotext, char *acltext));
+LOCAL	BOOL	acl_add_ids	__PR((char *name, char *infotext, char *acltext));
 #endif
 
 #ifdef	HAVE_SUN_ACL
 LOCAL	char	*acl_add_ids	__PR((char *dst, char *from, char *end, int *sizep));
 #endif
 
-LOCAL	char	*base_acl	__PR((int mode));
+LOCAL	char	*base_acl	__PR((mode_t mode));
 LOCAL	void	acl_check_ids	__PR((char *acltext, char *infotext));
 
 
 #ifdef HAVE_POSIX_ACL
 
+#define	DID_OPT_ACL
+EXPORT void
+opt_acl()
+{
+	printf(" acl");
+}
+
 /*
- * Get the access control list for a file and convert it into the format used by star.
+ * Get the access control list for a file and convert it into the format
+ * used by star.
  */
 EXPORT BOOL
 get_acls(info)
@@ -116,7 +152,7 @@ get_acls(info)
 	if (is_symlink(info))
 		return (TRUE);
 
-	if (!acl_to_info(info->f_name, ACL_TYPE_ACCESS, acl_access_text))
+	if (!acl_to_info(info->f_sname, ACL_TYPE_ACCESS, acl_access_text))
 		return (FALSE);
 	if (*acl_access_text != '\0') {
 		info->f_xflags |= XF_ACL_ACCESS;
@@ -124,7 +160,7 @@ get_acls(info)
 	}
 	if (!is_dir(info))
 		return (TRUE);
-	if (!acl_to_info(info->f_name, ACL_TYPE_DEFAULT, acl_default_text))
+	if (!acl_to_info(info->f_sname, ACL_TYPE_DEFAULT, acl_default_text))
 		return (FALSE);
 	if (*acl_default_text != '\0') {
 		info->f_xflags |= XF_ACL_DEFAULT;
@@ -157,9 +193,13 @@ acl_to_info(name, type, acltext)
 		if (err == ENOSYS)
 			return (TRUE);
 #endif
-		errmsg("Cannot get %sACL for '%s'.\n",
-			type==ACL_TYPE_DEFAULT?"default ":"", name);
-		xstats.s_getaclerrs++;
+		if (!errhidden(E_GETACL, name)) {
+			if (!errwarnonly(E_GETACL, name))
+				xstats.s_getaclerrs++;
+			errmsg("Cannot get %sACL for '%s'.\n",
+				type == ACL_TYPE_DEFAULT ? "default ":"", name);
+			(void) errabort(E_GETACL, name, TRUE);
+		}
 		return (FALSE);
 	}
 	seterrno(0);
@@ -168,9 +208,14 @@ acl_to_info(name, type, acltext)
 	if (text == NULL) {
 		if (geterrno() == 0)
 			seterrno(EX_BAD);
-		errmsg("Cannot convert %sACL entries to text for '%s'.\n",
-			type==ACL_TYPE_DEFAULT?"default ":"", name);
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, name)) {
+			if (!errwarnonly(E_BADACL, name))
+				xstats.s_badacl++;
+			errmsg(
+			    "Cannot convert %sACL entries to text for '%s'.\n",
+				type == ACL_TYPE_DEFAULT ? "default ":"", name);
+			(void) errabort(E_BADACL, name, TRUE);
+		}
 		return (FALSE);
 	}
 
@@ -182,7 +227,7 @@ acl_to_info(name, type, acltext)
 	/* remove comment fields */
 	c = text;
 	while ((c = strchr(c, '#')) != NULL) {
-		char *d = c;
+		char *d = c, *e;
 
 		while (c > text && strchr(" \t\r", *(c-1)))
 			c--;
@@ -193,10 +238,11 @@ acl_to_info(name, type, acltext)
 		}
 		while (*d && *d != '\n')
 			d++;
-		while ((*c++ = *d++) != '\0')
+		e = c;
+		while ((*e++ = *d++) != '\0')
 			;
 	}
-	
+
 	/* count fields */
 	for (c = text; *c != '\0'; c++) {
 		if (*c == '\n') {
@@ -204,8 +250,9 @@ acl_to_info(name, type, acltext)
 			entries++;
 		}
 	}
-	if (entries > 3) { /* > 4 on Solaris? */
-		if (!acl_add_ids(acltext, text)) {
+	if ((entries > 3) || /* > 4 on Solaris? */
+	    (type == ACL_TYPE_DEFAULT && entries >= 3)) {
+		if (!acl_add_ids(name, acltext, text)) {
 			acl_free((acl_t)text);
 			return (FALSE);
 		}
@@ -223,14 +270,16 @@ acl_to_info(name, type, acltext)
 }
 
 LOCAL BOOL
-acl_add_ids(infotext, acltext)
+acl_add_ids(name, infotext, acltext)
+	char	*name;
 	char	*infotext;
 	char	*acltext;
 {
 	int	size = PATH_MAX;
 	int	len;
 	char	*token;
-	Ulong	id;
+	uid_t	uid;
+	gid_t	gid;
 
 	/*
 	 * Add final nul to guarantee that the string is nul terminated.
@@ -245,37 +294,37 @@ acl_add_ids(infotext, acltext)
 		if (size < 0)
 			size = 0;
 
-		if (!strncmp(token, "user:", 5) &&
-		    !strchr(":,\n\r", token[5])) {
+		if (strncmp(token, "user:", 5) == 0 &&
+		    strchr(":,\n\r", token[5]) == NULL) {
 			char *username = &token[5], *c = username+1;
 
-			while (!strchr(":,\n\r", *c))
+			while (strchr(":,\n\r", *c) == NULL)
 				c++;
 			*c = '\0';
 			/* check for all-numeric user name */
 			while (c > username && isdigit(*(c-1)))
 				c--;
 			if (c > username &&
-			    uidname(username, c-username, &id)) {
+			    ic_uidname(username, c-username, &uid)) {
 				len = js_snprintf(infotext, size,
-					":%ld", id);
+					":%lld", (Llong)uid);
 				infotext += len;
 				size -= len;
 			}
-		} else if (!strncmp(token, "group:", 6) &&
-		           !strchr(":,\n\r", token[6])) {
+		} else if (strncmp(token, "group:", 6) == 0 &&
+			    strchr(":,\n\r", token[6]) == NULL) {
 			char *groupname = &token[6], *c = groupname+1;
 
-			while (!strchr(":,\n\r", *c))
+			while (strchr(":,\n\r", *c) == NULL)
 				c++;
 			*c = '\0';
 			/* check for all-numeric group name */
 			while (c > groupname && isdigit(*(c-1)))
 				c--;
 			if (c > groupname &&
-			    gidname(groupname, c-groupname, &id)) {
+			    ic_gidname(groupname, c-groupname, &gid)) {
 				len = js_snprintf(infotext, size,
-					":%ld", id);
+					":%lld", (Llong)gid);
 				infotext += len;
 				size -= len;
 			}
@@ -284,14 +333,20 @@ acl_add_ids(infotext, acltext)
 			*infotext++ = ',';
 			size--;
 		}
-		
+
 		token = strtok(NULL, ",\n\r");
 	}
 	if (size >= 0) {
 		*(--infotext) = '\0';
 	} else {
-		errmsgno(EX_BAD, "Cannot convert ACL entries (string too long).\n");
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, name)) {
+			if (!errwarnonly(E_BADACL, name))
+				xstats.s_badacl++;
+			errmsgno(EX_BAD,
+			"Cannot convert ACL entries (string too long) for '%s'.\n",
+						name);
+			(void) errabort(E_BADACL, name, TRUE);
+		}
 		return (FALSE);
 	}
 	return (TRUE);
@@ -316,25 +371,33 @@ set_acls(info)
 		strcpy(acltext,  base_acl(info->f_mode));
 	}
 	if ((acl = acl_from_text(acltext)) == NULL) {
-		errmsg("Cannot convert ACL '%s' to internal format for '%s'.\n",
-		       acltext, info->f_name);
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, info->f_name)) {
+			if (!errwarnonly(E_BADACL, info->f_name))
+				xstats.s_badacl++;
+			errmsg("Cannot convert ACL '%s' to internal format for '%s'.\n",
+					acltext, info->f_name);
+			(void) errabort(E_BADACL, info->f_name, TRUE);
+		}
 	} else {
 		if (acl_set_file(info->f_name, ACL_TYPE_ACCESS, acl) < 0) {
 			/*
 			 * XXX What should we do if errno is ENOTSUP/ENOSYS?
-			 */ 
-			errmsg("Cannot set ACL '%s' for '%s'.\n",
-				acltext, info->f_name);
-			xstats.s_setacl++;
+			 */
+			if (!errhidden(E_SETACL, info->f_name)) {
+				if (!errwarnonly(E_SETACL, info->f_name))
+					xstats.s_setacl++;
+				errmsg("Cannot set ACL '%s' for '%s'.\n",
+					acltext, info->f_name);
+				(void) errabort(E_SETACL, info->f_name, TRUE);
+			}
 
 			/* Fall back to chmod */
 /* XXX chmod has already been done! */
-/*			chmod(info->f_name, (int)info->f_mode);*/
+/*			chmod(info->f_name, info->f_mode);*/
 		}
 		acl_free(acl);
 	}
-	
+
 	/*
 	 * Only directories can have Default ACLs
 	 */
@@ -352,25 +415,38 @@ set_acls(info)
 		if (acl_delete_def_file(info->f_name) < 0) {
 			/*
 			 * XXX What should we do if errno is ENOTSUP/ENOSYS?
-			 */ 
-			errmsg("Cannot remove default ACL from '%s'.\n", info->f_name);
-			xstats.s_setacl++;
+			 */
+			if (!errhidden(E_SETACL, info->f_name)) {
+				if (!errwarnonly(E_SETACL, info->f_name))
+					xstats.s_setacl++;
+				errmsg("Cannot remove default ACL from '%s'.\n",
+								info->f_name);
+				(void) errabort(E_SETACL, info->f_name, TRUE);
+			}
 		}
 		return;
 #endif
 	}
 	if ((acl = acl_from_text(acltext)) == NULL) {
-		errmsg("Cannot convert default ACL '%s' to internal format for '%s'.\n",
-		       acltext, info->f_name);
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, info->f_name)) {
+			if (!errwarnonly(E_BADACL, info->f_name))
+				xstats.s_badacl++;
+			errmsg("Cannot convert default ACL '%s' to internal format for '%s'.\n",
+					acltext, info->f_name);
+			(void) errabort(E_BADACL, info->f_name, TRUE);
+		}
 	} else {
 		if (acl_set_file(info->f_name, ACL_TYPE_DEFAULT, acl) < 0) {
 			/*
 			 * XXX What should we do if errno is ENOTSUP/ENOSYS?
-			 */ 
-			errmsg("Cannot set default ACL '%s' for '%s'.\n",
-				acltext, info->f_name);
-			xstats.s_setacl++;
+			 */
+			if (!errhidden(E_SETACL, info->f_name)) {
+				if (!errwarnonly(E_SETACL, info->f_name))
+					xstats.s_setacl++;
+				errmsg("Cannot set default ACL '%s' for '%s'.\n",
+					acltext, info->f_name);
+				(void) errabort(E_SETACL, info->f_name, TRUE);
+			}
 		}
 		acl_free(acl);
 	}
@@ -380,22 +456,30 @@ set_acls(info)
 
 #ifdef	HAVE_SUN_ACL	/* Solaris */
 
+#define	DID_OPT_ACL
+EXPORT void
+opt_acl()
+{
+	printf(" acl");
+}
+
 /*
- * Get the access control list for a file and convert it into the format used by star.
+ * Get the access control list for a file and convert it into the format
+ * used by star.
  */
 EXPORT BOOL
 get_acls(info)
 	register FINFO	*info;
 {
-		 int		aclcount;
-		 aclent_t	*aclp;
+		int		aclcount;
+		aclent_t	*aclp;
 	register char		*acltext;
 	register char		*ap;
 	register char		*dp;
 	register char		*cp;
 	register char		*ep;
-		 int		asize;
-		 int		dsize;
+		int		asize;
+		int		dsize;
 
 	info->f_acl_access = NULL;
 	info->f_acl_default = NULL;
@@ -406,11 +490,24 @@ get_acls(info)
 	if (is_symlink(info))
 		return (TRUE);
 
-	if ((aclcount = acl(info->f_name, GETACLCNT, 0, NULL)) < 0) {
-		errmsg("Cannot get ACL count for '%s'.\n", info->f_name);
-		xstats.s_getaclerrs++;
+#ifdef	HAVE_ST_ACLCNT
+	aclcount = info->f_aclcnt;
+#else
+	if ((aclcount = acl(info->f_sname, GETACLCNT, 0, NULL)) < 0) {
+#ifdef	ENOSYS
+		if (geterrno() == ENOSYS)
+			return (TRUE);
+#endif
+		if (!errhidden(E_GETACL, info->f_name)) {
+			if (!errwarnonly(E_GETACL, info->f_name))
+				xstats.s_getaclerrs++;
+			errmsg("Cannot get ACL count for '%s'.\n",
+							info->f_name);
+			(void) errabort(E_GETACL, info->f_name, TRUE);
+		}
 		return (FALSE);
 	}
+#endif
 #ifdef	ACL_DEBUG
 	error("'%s' acl count %d\n", info->f_name, aclcount);
 #endif
@@ -422,14 +519,24 @@ get_acls(info)
 		 */
 		return (TRUE);
 	}
-	if ((aclp = (aclent_t *)malloc(sizeof(aclent_t) * aclcount)) == NULL) {
-		errmsg("Cannot malloc ACL buffer for '%s'.\n", info->f_name);
-		xstats.s_getaclerrs++;
+	if ((aclp = (aclent_t *)malloc(sizeof (aclent_t) * aclcount)) == NULL) {
+		if (!errhidden(E_GETACL, info->f_name)) {
+			if (!errwarnonly(E_GETACL, info->f_name))
+				xstats.s_getaclerrs++;
+			errmsg("Cannot malloc ACL buffer for '%s'.\n",
+							info->f_name);
+			(void) errabort(E_GETACL, info->f_name, TRUE);
+		}
 		return (FALSE);
 	}
-	if (acl(info->f_name, GETACL, aclcount, aclp) < 0) {
-		errmsg("Cannot get ACL entries for '%s'.\n", info->f_name);
-		xstats.s_getaclerrs++;
+	if (acl(info->f_sname, GETACL, aclcount, aclp) < 0) {
+		if (!errhidden(E_GETACL, info->f_name)) {
+			if (!errwarnonly(E_GETACL, info->f_name))
+				xstats.s_getaclerrs++;
+			errmsg("Cannot get ACL entries for '%s'.\n",
+							info->f_name);
+			(void) errabort(E_GETACL, info->f_name, TRUE);
+		}
 		return (FALSE);
 	}
 	seterrno(0);
@@ -438,8 +545,13 @@ get_acls(info)
 	if (acltext == NULL) {
 		if (geterrno() == 0)
 			seterrno(EX_BAD);
-		errmsg("Cannot convert ACL entries to text for '%s'.\n", info->f_name);
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, info->f_name)) {
+			if (!errwarnonly(E_BADACL, info->f_name))
+				xstats.s_badacl++;
+			errmsg("Cannot convert ACL entries to text for '%s'.\n",
+								info->f_name);
+			(void) errabort(E_BADACL, info->f_name, TRUE);
+		}
 		return (FALSE);
 	}
 #ifdef	ACL_DEBUG
@@ -457,7 +569,7 @@ get_acls(info)
 		ep = strchr(cp, ',');
 		if (ep == NULL)
 			ep = strchr(cp, '\0');
-			
+
 		if (*cp == 'd' && strncmp(cp, "default", 7) == 0) {
 			cp += 7;
 			dp = acl_add_ids(dp, cp, ep, &dsize);
@@ -472,8 +584,13 @@ get_acls(info)
 	if (ap == NULL || dp == NULL) {
 		acl_access_text[0] = '\0';
 		acl_default_text[0] = '\0';
-		errmsgno(EX_BAD, "Cannot convert ACL entries (string too long).\n");
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, info->f_name)) {
+			if (!errwarnonly(E_BADACL, info->f_name))
+				xstats.s_badacl++;
+			errmsgno(EX_BAD, "Cannot convert ACL entries (string too long) for '%s'.\n",
+							info->f_name);
+			(void) errabort(E_BADACL, info->f_name, TRUE);
+		}
 		return (FALSE);
 	}
 
@@ -521,7 +638,8 @@ acl_add_ids(dst, from, end, sizep)
 	register char	*np = dst;
 	register int	size = *sizep;
 	register int	amt;
-		 Ulong	id;
+		uid_t	uid;
+		gid_t	gid;
 
 	if (cp[0] == 'u' &&
 	    strncmp(cp, "user:", 5) == 0) {
@@ -538,13 +656,13 @@ acl_add_ids(dst, from, end, sizep)
 		if (ep)
 			*ep = '\0';
 		if (*cp) {
-			if (uidname(cp, 1000, &id)) {
+			if (ic_uidname(cp, 1000, &uid)) {
 				if (np[-1] == ',') {
 					--np;
 					size++;
 				}
 				amt = js_snprintf(np, size,
-					":%ld,", id);
+					":%lld,", (Llong)uid);
 				np += amt;
 				size -= amt;
 			}
@@ -553,7 +671,7 @@ acl_add_ids(dst, from, end, sizep)
 			*ep = ':';
 
 	} else if (cp[0] == 'g' &&
-	     strncmp(cp, "group:", 6) == 0) {
+	    strncmp(cp, "group:", 6) == 0) {
 		if (size <= (ep - cp +1)) {
 			*sizep = 0;
 			return (NULL);
@@ -567,13 +685,13 @@ acl_add_ids(dst, from, end, sizep)
 		if (ep)
 			*ep = '\0';
 		if (*cp) {
-			if (gidname(cp, 1000, &id)) {
+			if (ic_gidname(cp, 1000, &gid)) {
 				if (np[-1] == ',') {
 					--np;
 					size++;
 				}
 				amt = js_snprintf(np, size,
-					":%ld,", id);
+					":%lld,", (Llong)gid);
 				np += amt;
 				size -= amt;
 			}
@@ -639,6 +757,7 @@ set_acls(info)
 	aclent_t	*aclp;
 	char		acltext[PATH_MAX+1];
 	char		aclbuf[8192];
+	BOOL		no_acl = FALSE;
 
 	aclbuf[0] = '\0';
 	if (info->f_xflags & XF_ACL_ACCESS) {
@@ -685,23 +804,46 @@ set_acls(info)
 		 * We may need to delete an inherited ACL.
 		 */
 		strcpy(aclbuf, base_acl(info->f_mode));
+		no_acl = TRUE;
+	} else if (streql(aclbuf, base_acl(info->f_mode))) {
+		no_acl = TRUE;
 	}
 
 	seterrno(0);
 	if ((aclp = aclfromtext(aclbuf, &aclcount)) == NULL) {
 		if (geterrno() == 0)
 			seterrno(EX_BAD);
-		errmsg("Cannot convert ACL '%s' to internal format for '%s'.\n",
-		       aclbuf, info->f_name);
-		xstats.s_badacl++;
+		if (!errhidden(E_BADACL, info->f_name)) {
+			if (!errwarnonly(E_BADACL, info->f_name))
+				xstats.s_badacl++;
+			errmsg("Cannot convert ACL '%s' to internal format for '%s'.\n",
+				aclbuf, info->f_name);
+			(void) errabort(E_BADACL, info->f_name, TRUE);
+		}
 	} else {
 		if (acl(info->f_name, SETACL, aclcount, aclp) < 0) {
+			BOOL	no_error = FALSE;
 			/*
 			 * XXX What should we do if errno is ENOSYS?
-			 */ 
-			errmsg("Cannot set ACL '%s' for '%s'.\n",
-				aclbuf, info->f_name);
-			xstats.s_setacl++;
+			 */
+			if (no_acl) {
+				int	aclcnt;
+
+				/*
+				 * This should catch the ENOSYS case which
+				 * happens e.g. if the target is a socket.
+				 */
+				aclcnt = acl(info->f_name, GETACLCNT, 0, NULL);
+				if (aclcnt <= MIN_ACL_ENTRIES)
+					no_error = TRUE;
+			}
+			if (!no_error && !errhidden(E_SETACL, info->f_name)) {
+				if (!errwarnonly(E_SETACL, info->f_name))
+				errmsg("Cannot set ACL '%s' for '%s'.\n",
+					aclbuf, info->f_name);
+					xstats.s_setacl++;
+				(void) errabort(E_SETACL, info->f_name, TRUE);
+			}
 		}
 		free(aclp);
 	}
@@ -709,13 +851,16 @@ set_acls(info)
 
 #endif	/* HAVE_SUN_ACL Solaris */
 
-
 /*
  * Convert UNIX standard mode bits into base ACL
  */
 LOCAL char *
+#ifdef	PROTOTYPES
+base_acl(mode_t mode)
+#else
 base_acl(mode)
-	int	mode;
+	mode_t	mode;
+#endif
 {
 	static char _acltxt[] = "user::***,group::***,other::***";
 
@@ -753,18 +898,28 @@ acl_check_ids(acltext, infotext)
 		return;
 
 	while (token) {
-
-		if (!strncmp(token, "user:", 5) &&
-		    !strchr(":,", token[5])) {
-			char *username = &token[5], *c = username+1;
+		char *x = strchr(token, '#');
+		if (x != NULL) {
+			/*
+			 * Cut off any "[ \t]#" at the end of the token.
+			 */
+			while (x > token+1 && strchr(" \t", *(x-1)) != NULL)
+				x--;
+			*x = '\0';
+		}
+		if (strncmp(token, "user:", 5) == 0 &&
+		    strchr(":,", token[5]) == NULL) {
+			char *username = &token[5], *c = username;
 			char *perms, *auid;
-			Ulong dummy;
+			uid_t	udummy;
 			/* uidname does not check for NULL! */
 
 			/* check for damaged user names with spaces */
 			if (strchr(username, ':') == NULL) {
-				/* Looks like a damaged user name that had
-				   spaces in it, like "Joe,User". Repair. */
+				/*
+				 * Looks like a damaged user name that had
+				 * spaces in it, like "Joe,User". Repair.
+				 */
 				char *unexpected_sep = strchr(username, '\0');
 
 				if (strtok(NULL, ",")) {
@@ -774,21 +929,24 @@ acl_check_ids(acltext, infotext)
 			}
 
 			/* user name */
-			while (!strchr(":,", *c))
+			if (*username != ':') {
 				c++;
+				while (strchr(":,", *c) == NULL)
+					c++;
+			}
 			if (*c)
 				*c++ = '\0';
 
 			/* permissions */
 			perms = c;
-			while (!strchr(":,", *c))
+			while (strchr(":,", *c) == NULL)
 				c++;
 			if (*c)
 				*c++ = '\0';
 
 			/* identifier */
 			auid = c;
-			while (!strchr(":,", *c))
+			while (strchr(":,", *c) == NULL)
 				c++;
 			if (*c)
 				*c++ = '\0';
@@ -800,23 +958,25 @@ acl_check_ids(acltext, infotext)
 			 * XXX "joeuser" would be recognized as identical.
 			 */
 			if (*auid && (numeric ||
-			    !uidname(username, strlen(username)+1, &dummy)))
+			    !ic_uidname(username, strlen(username)+1, &udummy)))
 				username = auid;
 			js_snprintf(entry_buffer, PATH_MAX, "user:%s:%s",
 				username, perms);
 			token = entry_buffer;
 
-		} else if (!strncmp(token, "group:", 6) &&
-		    !strchr(":,", token[6])) {
-			char *groupname = &token[6], *c = groupname+1;
+		} else if (strncmp(token, "group:", 6) == 0 &&
+		    strchr(",", token[6]) == NULL) {
+			char *groupname = &token[6], *c = groupname;
 			char *perms, *agid;
-			Ulong dummy;
+			gid_t	gdummy;
 			/* gidname does not check for NULL! */
 
 			/* check for damaged group names with spaces */
 			if (strchr(groupname, ':') == NULL) {
-				/* Looks like a damaged group name that had
-				   spaces in it, like "Domain,Users". Repair. */
+				/*
+				 * Looks like a damaged group name that had
+				 * spaces in it, like "Domain,Users". Repair.
+				 */
 				char *unexpected_sep = strchr(groupname, '\0');
 
 				if (strtok(NULL, ",")) {
@@ -826,25 +986,28 @@ acl_check_ids(acltext, infotext)
 			}
 
 			/* group name */
-			while (!strchr(":,", *c))
+			if (*groupname != ':') {
 				c++;
+				while (strchr(":,", *c) == NULL)
+					c++;
+			}
 			if (*c)
 				*c++ = '\0';
 
 			/* permissions */
 			perms = c;
-			while (!strchr(":,", *c))
+			while (strchr(":,", *c) == NULL)
 				c++;
 			if (*c)
 				*c++ = '\0';
 
 			/* identifier */
 			agid = c;
-			while (!strchr(":,", *c))
+			while (strchr(":,", *c) == NULL)
 				c++;
 			if (*c)
 				*c++ = '\0';
-			
+
 			/*
 			 * XXX We use strlen(groupname)+1 to tell gidname not
 			 * XXX to stop comparing before the end of the
@@ -852,26 +1015,28 @@ acl_check_ids(acltext, infotext)
 			 * XXX "joeuser" would be compared as identical.
 			 */
 			if (*agid && (numeric ||
-			    !gidname(groupname, strlen(groupname)+1, &dummy)))
+			    !ic_gidname(groupname, strlen(groupname)+1, &gdummy)))
 				groupname = agid;
 			js_snprintf(entry_buffer, PATH_MAX, "group:%s:%s",
 				groupname, perms);
 			token = entry_buffer;
-		} else if (!strncmp(token, "#effective:", 11)) {
-			/* Older versions of star didn't trim off
-			 * effective rights comments, so ignore them
-			 * when they are found in archives.
-			 */
-			goto skip_malformed_entry;
 		}
-		strcpy(acltext, token);
-		acltext += strlen(token);
-		*acltext++ = ',';
+		if (*token != '\0') {
+			strcpy(acltext, token);
+			acltext += strlen(token);
+			*acltext++ = ',';
+		}
 
-	skip_malformed_entry:
 		token = strtok(NULL, ",");
 	}
 	*(--acltext) = '\0';
 }
 
 #endif  /* USE_ACL */
+
+#ifndef	DID_OPT_ACL
+EXPORT void
+opt_acl()
+{
+}
+#endif
