@@ -1,7 +1,8 @@
-/* @(#)props.c	1.27 02/06/17 Copyright 1994 J. Schilling */
+/* @(#)props.c	1.54 09/07/11 Copyright 1994-2009 J. Schilling */
+#include <schily/mconfig.h>
 #ifndef lint
-static	char sccsid[] =
-	"@(#)props.c	1.27 02/06/17 Copyright 1994 J. Schilling";
+static	UConst char sccsid[] =
+	"@(#)props.c	1.54 09/07/11 Copyright 1994-2009 J. Schilling";
 #endif
 /*
  *	Set up properties for different archive types
@@ -16,32 +17,27 @@ static	char sccsid[] =
  *	pr_flags/pr_nflags or the fields pr_xftypetab[]/pr_typeflagtab[]
  *	take care of possible problems due to this fact.
  *
- *	Copyright (c) 1994 J. Schilling
+ *	Copyright (c) 1994-2009 J. Schilling
  */
 /*
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
- * any later version.
+ * The contents of this file are subject to the terms of the
+ * Common Development and Distribution License, Version 1.0 only
+ * (the "License").  You may not use this file except in compliance
+ * with the License.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * See the file CDDL.Schily.txt in this distribution for details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; see the file COPYING.  If not, write to
- * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+ * When distributing Covered Code, include this CDDL HEADER in each
+ * file and include the License file CDDL.Schily.txt from this distribution.
  */
 
-#include <mconfig.h>
-#include <stdio.h>
+#include <schily/stdio.h>
 #include "star.h"
 #include "props.h"
 #include "table.h"
 #include "diff.h"
-#include <standard.h>
-#include <schily.h>
+#include <schily/standard.h>
+#include <schily/schily.h>
 #include "starsubs.h"
 
 extern	BOOL	debug;
@@ -57,16 +53,19 @@ EXPORT void
 setprops(htype)
 	long	htype;
 {
-	fillbytes(props.pr_typeflagtab, sizeof(props.pr_typeflagtab), '\0');
+	fillbytes(props.pr_typeflagtab, sizeof (props.pr_typeflagtab), '\0');
 
 	switch (H_TYPE(htype)) {
 
 	case H_STAR:				/* Old star format (1985)    */
 		props.pr_maxsize = 0;
+		props.pr_hdrsize = TAR_HDRSZ;
 		props.pr_flags = PR_LOCAL_STAR|PR_SPARSE|PR_VOLHDR|PR_BASE256;
-		props.pr_xdflags = 0;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
 		props.pr_fillc = ' ';		/* Use old tar octal format  */
 		props.pr_xc    = 'x';		/* Use POSIX.1-2001 x-hdr    */
+		props.pr_pad   = 0;
 		props.pr_diffmask = 0L;
 		props.pr_nflags =
 			PR_POSIX_SPLIT|PR_PREFIX_REUSED|PR_LONG_NAMES;
@@ -79,30 +78,37 @@ setprops(htype)
 		/*
 		 * The STAR format extensions from 1986 archive nearly any UNIX file type
 		 */
-		movebytes(xtstar_tab, props.pr_xftypetab, sizeof(props.pr_xftypetab));
+		movebytes(xtstar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
 
 		/*
 		 * The STAR format is pre-POSIX but supports many POSIX/GNU
 		 * extensions.
 		 */
 		prsettypeflags("01234567gxIKLMSV", TF_VALIDTYPE);
-		prsettypeflags("gxKLM",            TF_XHEADERS);
+		prsettypeflags("gxKL",		   TF_XHEADERS);
 		break;
 
 	case H_XSTAR:				/* ext P-1988 format (1994)  */
 	case H_XUSTAR:				/* ext ^ no "tar" sig (1998) */
 	case H_EXUSTAR:				/* ext P-2001 format (2001)  */
 		props.pr_maxsize = 0;
+		props.pr_hdrsize = TAR_HDRSZ;
 		props.pr_flags =
-			PR_POSIX_OCTAL|PR_LOCAL_STAR|PR_SPARSE|PR_VOLHDR|PR_BASE256;
-		if (H_TYPE(htype) == H_XUSTAR || H_TYPE(htype) == H_EXUSTAR)
-			props.pr_flags |= PR_XHDR;
-		props.pr_xdflags = 0;
-		if (H_TYPE(htype) == H_EXUSTAR)
-			props.pr_xdflags |= (XF_ATIME|XF_CTIME|XF_MTIME);
-		
+			PR_POSIX_OCTAL|PR_LOCAL_STAR|PR_SPARSE|
+			PR_MULTIVOL|PR_VOLHDR|PR_BASE256;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
+		if (H_TYPE(htype) == H_XUSTAR || H_TYPE(htype) == H_EXUSTAR) {
+			props.pr_flags |= (PR_XHDR|PR_VU_XHDR);
+			props.pr_xhmask = ~0;
+		}
+		if (H_TYPE(htype) == H_EXUSTAR) {
+			props.pr_flags |= PR_LINK_DATA;
+			props.pr_xhdflags |= (XF_ATIME|XF_CTIME|XF_MTIME);
+		}
 		props.pr_fillc = '0';		/* Use ustar octal format    */
 		props.pr_xc    = 'x';		/* Use POSIX.1-2001 x-hdr    */
+		props.pr_pad   = 0;
 		props.pr_diffmask = 0L;
 		props.pr_nflags =
 			PR_POSIX_SPLIT|PR_PREFIX_REUSED|PR_LONG_NAMES;
@@ -118,7 +124,7 @@ setprops(htype)
 		 * sparse files as well as
 		 * meta files (regular files without archived content).
 		 */
-		movebytes(xtustar_tab, props.pr_xftypetab, sizeof(props.pr_xftypetab));
+		movebytes(xtustar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
 		props.pr_xftypetab[XT_SPARSE] = 1;
 		props.pr_xftypetab[XT_META] = 1;
 
@@ -131,37 +137,47 @@ setprops(htype)
 		 */
 /*		if (H_TYPE(htype) == H_EXUSTAR)*/
 		if (dodump && H_TYPE(htype) == H_EXUSTAR)
-			movebytes(xtexustar_tab, props.pr_xftypetab, sizeof(props.pr_xftypetab));
+			movebytes(xtexustar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
 
 		/*
 		 * The X-STAR format family is POSIX with extensions.
 		 */
 		prsettypeflags("01234567gxIKLMSV", TF_VALIDTYPE);
-		prsettypeflags("gxKLM",            TF_XHEADERS);
+		prsettypeflags("gxKL",		   TF_XHEADERS);
 		break;
 
 	case H_PAX:				/* ieee 1003.1-2001 ext ustar*/
 	case H_USTAR:				/* ieee 1003.1-1988 ustar    */
 	case H_SUNTAR:				/* Sun's tar from Solaris 7-9*/
 		props.pr_maxsize = MAXOCTAL11;
+		props.pr_hdrsize = TAR_HDRSZ;
 		props.pr_flags = PR_POSIX_OCTAL;
 		if (H_TYPE(htype) == H_PAX || H_TYPE(htype) == H_SUNTAR) {
 			props.pr_maxsize = 0;
 			props.pr_flags |= PR_XHDR;
 		}
-		props.pr_xdflags = 0;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
+		if (H_TYPE(htype) == H_PAX)
+			props.pr_xhmask = XF_POSIX;
 		props.pr_fillc = '0';		/* Use ustar octal format    */
 		props.pr_xc    = 'x';		/* Use POSIX.1-2001 x-hdr    */
+		props.pr_pad   = 0;
 		if (H_TYPE(htype) == H_SUNTAR) {
-			props.pr_xdflags = XF_MTIME;
+			props.pr_flags |= PR_VU_XHDR;
+			props.pr_xhdflags = XF_MTIME;
+			props.pr_xhmask =   XF_MTIME|XF_UID|XF_UNAME|
+					    XF_GID|XF_GNAME|
+					    XF_PATH|XF_LINKPATH|
+					    XF_SIZE|XF_DEVMAJOR|XF_DEVMINOR;
 			props.pr_xc    = 'X';	/* Use Sun Solaris  X-hdr    */
 		}
-		props.pr_diffmask = (D_ATIME|D_CTIME);
+		props.pr_diffmask = (D_SPARS|D_ATIME|D_CTIME);
 		if (H_TYPE(htype) == H_PAX) {
 			props.pr_diffmask = 0L;
 		}
 		props.pr_nflags = PR_POSIX_SPLIT;
-		props.pr_maxnamelen =  NAMSIZ;
+		props.pr_maxnamelen =  NAMSIZ + 1 + PFXSIZ;	/* 256 */
 		props.pr_maxlnamelen = NAMSIZ;
 		if (H_TYPE(htype) == H_PAX || H_TYPE(htype) == H_SUNTAR) {
 			props.pr_nflags |= PR_LONG_NAMES;
@@ -175,26 +191,29 @@ setprops(htype)
 		/*
 		 * USTAR is limited to 7 basic file types
 		 */
-		movebytes(xtustar_tab, props.pr_xftypetab, sizeof(props.pr_xftypetab));
+		movebytes(xtustar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
 
 		/*
 		 * Even the USTAR format should support POSIX extension headers.
 		 */
-		prsettypeflags("01234567gx", TF_VALIDTYPE);
-		prsettypeflags("gx",         TF_XHEADERS);
+		prsettypeflags("01234567gx",	TF_VALIDTYPE);
+		prsettypeflags("gx",		TF_XHEADERS);
 		if (H_TYPE(htype) == H_SUNTAR) {
-			prsettypeflags("01234567gxX", TF_VALIDTYPE);
-			prsettypeflags("gxX",         TF_XHEADERS);
+			prsettypeflags("01234567gxX",	TF_VALIDTYPE);
+			prsettypeflags("gxX",		TF_XHEADERS);
 		}
 		break;
 
-	case H_GNUTAR:				/* gnu tar format (1989)     */
+	case H_GNUTAR:				/* gnu tar format (1989)    */
 		props.pr_maxsize = 0;
+		props.pr_hdrsize = TAR_HDRSZ;
 		props.pr_flags =
 			PR_LOCAL_GNU|PR_SPARSE|PR_GNU_SPARSE_BUG|PR_VOLHDR|PR_BASE256;
-		props.pr_xdflags = 0;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
 		props.pr_fillc = ' ';		/* Use old tar octal format  */
 		props.pr_xc    = 'x';		/* Really ??? */
+		props.pr_pad   = 0;
 		props.pr_diffmask = 0L;
 		props.pr_nflags = PR_LONG_NAMES;
 		props.pr_maxnamelen =  PATH_MAX;
@@ -206,16 +225,26 @@ setprops(htype)
 		/*
 		 * GNU tar is limited to the USTAR file types + sparse files
 		 */
-		movebytes(xtustar_tab, props.pr_xftypetab, sizeof(props.pr_xftypetab));
+		movebytes(xtustar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
 		props.pr_xftypetab[XT_SPARSE] = 1;
 
 		prsettypeflags("01234567DKLMNSV", TF_VALIDTYPE);
-		prsettypeflags("KLM",             TF_XHEADERS);
+		prsettypeflags("KL",		  TF_XHEADERS);
 		break;
 
-	case H_TAR:				/* tar with unknown format   */
-	case H_OTAR:				/* tar old format (1978 ???) */
 	default:
+		errmsgno(EX_BAD, "setprops: unexpected header type %ld '%s' (%s).\n",
+						htype,
+						hdr_name(htype),
+						hdr_text(htype));
+		errmsgno(EX_BAD, "setprops: defaulting to '%s' (%s).\n",
+						hdr_name(H_OTAR),
+						hdr_text(H_OTAR));
+
+	case H_UNDEF:				/* from star main read mode  */
+	case H_TAR:				/* tar with unknown format   */
+	case H_V7TAR:				/* tar old UNIX V7 format    */
+	case H_OTAR:				/* tar old BSD format (1978) */
 		/*
 		 * Since the large file summit was in 1995, we may assume
 		 * that any large file aware TAR implementation must be POSIX
@@ -225,11 +254,15 @@ setprops(htype)
 		 * by non large file aware systems.
 		 */
 		props.pr_maxsize = MAXNONLARGEFILE;
+/*		props.pr_maxsize = MAXOCTAL11;*/
+		props.pr_hdrsize = TAR_HDRSZ;
 		props.pr_flags = 0;
-		props.pr_xdflags = 0;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
 		props.pr_fillc = ' ';		/* Use old tar octal format  */
 		props.pr_xc    = 'x';		/* Really ??? */
-		props.pr_diffmask = (D_ATIME|D_CTIME);
+		props.pr_pad   = 0;
+		props.pr_diffmask = (D_SPARS|D_ATIME|D_CTIME);
 		props.pr_nflags = PR_DUMB_EOF;
 		props.pr_maxnamelen =  NAMSIZ-1;
 		props.pr_maxlnamelen = NAMSIZ-1;
@@ -240,15 +273,131 @@ setprops(htype)
 		/*
 		 * Only a very limited set of file types (file,symlink,dir)
 		 */
-		movebytes(xttar_tab, props.pr_xftypetab, sizeof(props.pr_xftypetab));
+		if (H_TYPE(htype) == H_V7TAR)
+			movebytes(xtv7tar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
+		else
+			movebytes(xttar_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
 
 		/*
-		 * Old tar does no know about file types bejond "012" but the
-		 * USTAR basic format has been designed to be compatible with
-		 * the old tar format.
+		 * Old BSD tar does no know about file types bejond "012" but
+		 * the USTAR basic format has been designed to be compatible
+		 * with the old tar format.
+		 * OLD UNIX V7 tar does not know anything but plain files and
+		 * hard links. Note that we never see H_TYPE(htype) == H_V7TAR
+		 * in extract mode.
 		 */
-		prsettypeflags("01234567", TF_VALIDTYPE);
-		prsettypeflags("",         TF_XHEADERS);
+		if (H_TYPE(htype) == H_V7TAR)
+			prsettypeflags("01", TF_VALIDTYPE);
+		else
+			prsettypeflags("01234567", TF_VALIDTYPE);
+		prsettypeflags("",	   TF_XHEADERS);
+		break;
+
+	case H_BAR:				/* Sun bar format */
+		props.pr_maxsize = MAXNONLARGEFILE; /* bar is pre-largefile */
+		props.pr_hdrsize = BAR_HDRSZ;
+		props.pr_flags = 0;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
+		props.pr_fillc = ' ';		/* Use old tar octal format  */
+		props.pr_xc    = 'x';		/* Really ??? */
+		props.pr_pad   = 0;
+		props.pr_diffmask = (D_SPARS|D_ATIME|D_CTIME);
+		props.pr_nflags = PR_DUMB_EOF;
+		props.pr_maxnamelen =  PATH_MAX;
+		props.pr_maxlnamelen = PATH_MAX;
+		props.pr_maxsname =    NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxslname =   NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxprefix =   0;
+		props.pr_sparse_in_hdr = 0;
+		/*
+		 * BAR is limited to 7 basic file types
+		 */
+		movebytes(xtcpio_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
+		props.pr_xftypetab[XT_SOCK] = 0;
+		break;
+
+	case H_CPIO_BIN:
+		props.pr_maxsize = MAXNONLARGEFILE;
+		props.pr_hdrsize = CPIOBIN_HDRSZ;
+		props.pr_flags = PR_CPIO;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
+		props.pr_fillc = '0';
+		props.pr_xc    = 'x';		/* Not needed */
+		props.pr_pad   = 1;
+		props.pr_diffmask = (D_SPARS|D_ATIME|D_CTIME);
+		props.pr_nflags = PR_LONG_NAMES;
+		props.pr_maxnamelen =  256;
+		props.pr_maxlnamelen = PATH_MAX;
+		props.pr_maxsname =    NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxslname =   NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxprefix =   0;
+		props.pr_sparse_in_hdr = 0;
+		/*
+		 * CPIO is limited to 7 basic file types
+		 */
+		movebytes(xtcpio_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
+		props.pr_xftypetab[XT_SOCK] = 0;
+		break;
+
+	case H_CPIO_CHR:
+	case H_CPIO_ODC:
+		props.pr_maxsize = MAXOCTAL11;
+		props.pr_hdrsize = CPIOODC_HDRSZ;
+		props.pr_flags = PR_CPIO;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
+		props.pr_fillc = '0';
+		props.pr_xc    = 'x';		/* Not needed */
+		props.pr_pad   = 0;
+		props.pr_diffmask = (D_SPARS|D_ATIME|D_CTIME);
+		props.pr_nflags = PR_LONG_NAMES;
+		props.pr_maxnamelen =  PATH_MAX;
+		props.pr_maxlnamelen = PATH_MAX;
+		props.pr_maxsname =    NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxslname =   NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxprefix =   0;
+		props.pr_sparse_in_hdr = 0;
+		/*
+		 * CPIO is limited to 7 basic file types + sockets
+		 */
+		movebytes(xtcpio_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
+
+		if (H_TYPE(htype) == H_CPIO_ODC) {
+			/*
+			 * Be compatible to the useless SYSv cpio -Hodc
+			 * limitations.
+			 */
+			props.pr_maxnamelen =  255;
+			props.pr_xftypetab[XT_SOCK] = 0;
+		}
+		break;
+
+	case H_CPIO_ASC:
+	case H_CPIO_ACRC:
+		props.pr_maxsize = (unsigned long)0xFFFFFFFF;
+		props.pr_hdrsize = CPIOCRC_HDRSZ;
+		props.pr_flags = PR_CPIO | PR_SV_CPIO_LINKS;
+		props.pr_xhdflags = 0;
+		props.pr_xhmask = 0;
+		props.pr_fillc = '0';
+		props.pr_xc    = 'x';		/* Not needed */
+		props.pr_pad   = 3;
+		props.pr_diffmask = (D_SPARS|D_ATIME|D_CTIME);
+		props.pr_nflags = PR_LONG_NAMES;
+		props.pr_maxnamelen =  PATH_MAX;
+		props.pr_maxlnamelen = PATH_MAX;
+		props.pr_maxsname =    NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxslname =   NAMSIZ-1; /* XXX Really ??? */
+		props.pr_maxprefix =   0;
+		props.pr_sparse_in_hdr = 0;
+		/*
+		 * CPIO is limited to 7 basic file types
+		 */
+		movebytes(xtcpio_tab, props.pr_xftypetab, sizeof (props.pr_xftypetab));
+		props.pr_xftypetab[XT_SOCK] = 0;
+		break;
 	}
 	if (debug) printprops();
 }
@@ -257,13 +406,14 @@ EXPORT void
 printprops()
 {
 extern long hdrtype;
-	error("hdrtype:          %ld\n", hdrtype);
+	error("hdrtype:          %s (%ld)\n", hdr_name(hdrtype), hdrtype);
 	if (props.pr_maxsize)
 		error("pr_maxsize:       %llu\n", (Ullong)props.pr_maxsize);
 	else
 		error("pr_maxsize:       unlimited\n");
 	error("pr_flags:         0x%X\n", props.pr_flags);
-	error("pr_xdflags:       0x%X\n", props.pr_xdflags);
+	error("pr_xhdflags:      0x%X\n", props.pr_xhdflags);
+	error("pr_xhmask:        0x%X\n", props.pr_xhmask);
 	error("pr_fillc:         '%c'\n", props.pr_fillc);
 	prdiffopts(stderr, "pr_diffmask:      ", props.pr_diffmask);
 	error("pr_nflags:        0x%X\n", props.pr_nflags);
